import logging

from .handlers import ChatworkHandler, SlcakHandler, TelegramHandler
from .settings import LoggerSettings


def get_logger(env_file: str = ".env"):
    conf = LoggerSettings(_env_file=env_file)
    logger = logging.getLogger("")
    logger.setLevel(0)
    if conf.use_telegram:
        th = TelegramHandler(api_token=conf.telegram_token, chat_id=conf.telegram_chat_id)
        th.setLevel(conf.telegram_level)
        logger.addHandler(th)

    if conf.use_slack:
        sh = SlcakHandler(token=conf.slack_token, channel=conf.slack_channel)
        sh.setLevel(conf.slack_level)
        logger.addHandler(sh)

    if conf.use_chatwork:
        ch = ChatworkHandler(token=conf.chatwork_token, room_id=conf.chatwork_room_id)
        sh.setLevel(conf.chatwork_level)
        logger.addHandler(ch)

    return logger



import io
import logging
import traceback
from typing import Optional


def get_other_info(record: logging.LogRecord) -> Optional[str]:
    """stack_infoかexec_info(traceback)を取り出す

    Args:
        record (logging.LogRecord): レコード

    Returns:
        Optional[str]: traceback or exec_info or None
    """
    if record.exc_text is not None:
        return record.exc_text
    elif record.exc_info is not None:
        ei = record.exc_info
        with io.StringIO() as sio:
            traceback.print_exception(ei[0], ei[1], ei[2], None, sio)
            return sio.getvalue()
    elif record.stack_info is not None:
        return record.stack_info
    return None



import logging

from .request_backend import RequestBackend

rb = RequestBackend()


class HTTPHandler(logging.Handler):
    def __init__(self, url: str, headers: dict = {}, mode: str = "json", **kwargs):
        """これを継承して、各WebサービスにあったHanderを作成する

        Args:
            url (str): ログを送るサービスのエンドポイント
            headers (Optional[dict], optional): ヘッダー(認証用). Defaults to None.
            mode (Literal["json", "data"], optional): mode. Defaults to "json".
        """
        self.url = url
        self.headers = headers
        self.mode = mode
        self.rb = rb
        self.rb.init()

        super().__init__(**kwargs)

    def mapLogRecord(self, record: logging.LogRecord) -> dict:
        return {k: v for k, v in record.__dict__.items() if isinstance(v, str)}

    def emit(self, record: logging.LogRecord) -> None:
        kwargs = {
            "url": self.url,
            self.mode: self.mapLogRecord(record),
            "headers": self.headers,
        }
        self.rb.post(**kwargs)



import logging

from .._helper import get_other_info
from .web_handler import HTTPHandler


class SlcakHandler(HTTPHandler):
    def __init__(self, token: str, channel: str, **kwargs):
        self.channel = channel
        url = "https://slack.com/api/chat.postMessage"
        headers = {"Authorization": f"Bearer {token}"}
        super().__init__(url, headers, **kwargs)

    def mapLogRecord(self, record: logging.LogRecord) -> dict:
        loglevel_color = {
            "DEBUG": "gray",
            "INFO": "good",
            "WARNING": "warning",
            "ERROR": "#E91E63",
            "CRITICAL": "danger",
        }
        self.format(record)
        text = record.msg
        other_info = get_other_info(record)
        if other_info is not None:
            text += "\n" + f"```{other_info}```"

        content = dict(
            ts=record.created,
            text=text,
            color=loglevel_color.get(record.levelname, "white"),
            footer=f"{record.pathname}:{record.lineno} [{record.levelname}]",
        )

        return dict(attachments=[content], channel=self.channel)



import html
import logging
from datetime import datetime

from .._helper import get_other_info
from .web_handler import HTTPHandler


class TelegramHandler(HTTPHandler):
    def __init__(self, api_token: str, chat_id: int, **kwargs):
        self.chat_id = chat_id
        url = f"https://api.telegram.org/bot{api_token}/sendMessage"
        super().__init__(url, **kwargs)

    def mapLogRecord(self, record: logging.LogRecord) -> dict:
        level = TelegramStyled(record.levelname).b.u.s
        time = TelegramStyled(str(datetime.fromtimestamp(record.created))).i.b.u.s
        msg = record.getMessage()
        path = TelegramStyled(f"from: {record.pathname}:{record.lineno}").b.u.s
        other_info = get_other_info(record)
        text: str = level + " " + time + "\n" + path
        if msg != "":
            text += "\n" + msg
        if other_info is not None:
            text += "\n" + TelegramStyled(other_info).c.s

        return dict(chat_id=self.chat_id, parse_mode="HTML", text=text)


class TelegramStyled:
    def __init__(self, s: str):
        self.s = html.escape(s)

    @property
    def u(self) -> "TelegramStyled":
        self.s = f"<u>{self.s}</u>"
        return self

    @property
    def b(self) -> "TelegramStyled":
        self.s = f"<b>{self.s}</b>"
        return self

    @property
    def i(self) -> "TelegramStyled":
        self.s = f"<i>{self.s}</i>"
        return self

    @property
    def c(self) -> "TelegramStyled":
        self.s = f'<pre><code class="language-python">{self.s}</code></pre>'
        return self



import logging
from datetime import datetime

from .._helper import get_other_info
from .web_handler import HTTPHandler

cw_msg = """[info][title]{level_name} {time}
{path_name}:{line_no}
[/title]
{msg}{other_info}
[/info]
"""


def code(s: str) -> str:
    return f"[code]{s}[/code]"


class ChatworkHandler(HTTPHandler):
    def __init__(self, token: str, room_id: int, **kwargs):
        url = f"https://api.chatwork.com/v2/rooms/{room_id}/messages"
        headers = {"X-ChatWorkToken": token}
        super().__init__(url, headers, mode="data", **kwargs)

    def mapLogRecord(self, record: logging.LogRecord) -> dict:
        self.format(record)
        other_info = get_other_info(record)
        other_info = "" if other_info is None else "[hr]" + code(other_info)

        msg = cw_msg.format(
            level_name=record.levelname,
            time=str(datetime.fromtimestamp(record.created)),
            path_name=record.pathname,
            line_no=record.lineno,
            msg="" if record.msg == "" else code(record.msg),
            other_info=other_info,
        )

        return {"body": msg}



__version__ = "0.1.0"

from agarilog.log import get_logger


def reset_global_alog():
    global debug
    global info
    global warn
    global warning
    global error
    global critical
    default_logger = get_logger()
    debug = default_logger.debug
    info = default_logger.info
    warn = default_logger.warn
    warning = default_logger.warning
    error = default_logger.error
    critical = default_logger.critical


reset_global_alog()



"""メインスレッドとは違うところでQueueにたまったリクエストを送り続けるバックエンド機構
"""
import asyncio
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from typing import Dict, List, Optional, Tuple, Union

import aiohttp
from pydantic import BaseModel, HttpUrl, validator

from ._helper import register, to_thread


class Request(BaseModel):
    """Requestするためのデータセット

    Args:
        method (Literal["GET", "POST"]): HTTPメソッド
        url (HttpUrl): URL
        headers (Dict[str, str]): ヘッダー
        data (Optional[bytes]): データ
        json_ (Optional[dict]): jsonを送るときはこっち
        params (Optional[Union[List[Tuple[str, str]], Dict[str, str]]]): query string
    """

    method: str
    url: HttpUrl
    headers: Dict[str, str]
    data: Optional[Union[dict, bytes]]
    json_: Optional[dict]
    params: Optional[Union[List[Tuple[str, str]], Dict[str, str]]]

    @validator("method")
    def _valid_method(cls, v):
        if v not in ["GET", "POST"]:
            raise ValueError('method is "GET" or "POST"')
        return v


class RequestBackend:
    def __init__(self) -> None:
        self.flg_init = False
        register(self)  # 待機する処理を加える(IPythonと通常実行で処理を切り替える)

    def post(
        self,
        url: str,
        params: Optional[Union[List[Tuple[str, str]], Dict[str, str]]] = None,
        data: Optional[Union[dict, bytes]] = None,
        json: Optional[dict] = None,
        headers: Dict[str, str] = {},
    ):
        request = Request(
            method="POST", url=url, params=params, data=data, json_=json, headers=headers
        )
        self._put(request)

    def get(
        self,
        url: str,
        params: Optional[Union[List[Tuple[str, str]], Dict[str, str]]] = None,
        headers: Dict[str, str] = {},
    ):
        request = Request(method="GET", url=url, params=params, headers=headers)
        self._put(request)

    def init(self):
        if self.flg_init:
            return
        self.queue: Queue[Union[Request], str] = Queue()
        self.executor = ThreadPoolExecutor()
        self.future = self.executor.submit(asyncio.run, self._worker())
        self.flg_init = True

    def shutdown(self):
        if not self.flg_init:
            return
        self._put("STOP")
        self.future.result()
        self.executor.shutdown()
        self.flg_init = False

    def _put(self, item: Union[Request, str]):
        self.queue.put_nowait(item)

    async def _send_request(self, session: aiohttp.ClientSession, request: Request):
        method = self._get if request.method == "GET" else self._post
        await method(session, request)

    @staticmethod
    async def _get(sess: aiohttp.ClientSession, r: Request):
        obj = dict(url=r.url, params=r.params, headers=r.headers)
        async with sess.get(**obj) as res:
            res.raise_for_status()

    @staticmethod
    async def _post(sess: aiohttp.ClientSession, r: Request):
        obj = dict(url=str(r.url), params=r.params, data=r.data, json=r.json_, headers=r.headers)
        async with sess.post(**obj) as res:
            res.raise_for_status()

    async def _worker(self) -> bool:
        """QueueにたまったRequestを処理する奴

        Returns:
            bool: 完了したらTrueを返す
        """
        async with aiohttp.ClientSession() as session:
            tasks = []
            while True:
                request = await to_thread(self.queue.get)
                if request == "STOP":
                    break
                send_request = self._send_request(session, request)
                tasks.append(asyncio.create_task(send_request))
                tasks = [task for task in tasks if not task.done()]
            if len(tasks) != 0:
                await asyncio.wait(tasks)
        return True



import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .request_backend import RequestBackend

try:
    ip = eval("get_ipython()")

    def register(rb: "RequestBackend"):
        """Decorator that registers at post_execute. After its execution it
        unregisters itself for subsequent runs."""

        ip.events.register("pre_execute", rb.init)
        ip.events.register("post_execute", rb.shutdown)


except NameError:
    import atexit

    def register(rb: "RequestBackend"):
        rb.init()
        atexit.register(rb.shutdown)


async def to_thread(func, *args, **kwargs):
    with ThreadPoolExecutor() as executor:
        loop = asyncio.get_running_loop()
        coro = loop.run_in_executor(executor, func, *args, **kwargs)
        return await coro



from typing import Optional

from pydantic import BaseSettings, validator


class LoggerSettings(BaseSettings):
    telegram_token: Optional[str]
    telegram_chat_id: Optional[int]
    telegram_level: str = "WARNING"
    slack_token: Optional[str]
    slack_channel: Optional[str]
    slack_level: str = "WARNING"
    chatwork_token: Optional[str]
    chatwork_room_id: Optional[int]
    chatwork_level: str = "WARNING"

    @validator("telegram_level", "slack_level", "chatwork_level")
    def _valid_level(cls, v):
        levels = ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        if v not in levels:
            raise ValueError(
                f'{v} is not in ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]'
            )
        return v

    @property
    def use_telegram(self) -> bool:
        return self._check_none("telegram")

    @property
    def use_slack(self) -> bool:
        return self._check_none("slack")

    @property
    def use_chatwork(self) -> bool:
        return self._check_none("chatwork")

    def _check_none(self, pre_fix: str) -> bool:
        arr = [v for k, v in dict(self).items() if pre_fix in k]
        return all(arr)

    class Config:
        env_prefix = "log_"
        env_file = ".env"
        env_file_encoding = "utf-8"
